<?php
require_once('util.inc');
require_once('functions.inc');
require_once('config.inc');
require_once("alias-utils.inc");

global $g, $config, $ipqs;
if (!is_array($ipqs)) {
	$ipqs = array();
}

/**
 * ipqs_gbl is called to default all values need by the ipqs module.
 * 
 * @return void
 */
$ipqs['fnc']['global_init'] = function(){
    global $config, $g, $ipqs;
    $config = parse_config(true);

    //global auth.

    if(isset($ipqs['config']['gbl']['authorized'])        === false) $ipqs['config']['gbl']['authorized']       = "127.0.0.1";
    if(isset($ipqs['config']['oauth']['domain'])          === false) $ipqs['config']['oauth']['domain']         = '';
    if(isset($ipqs['config']['oauth']['access_token'])    === false) $ipqs['config']['oauth']['access_token']   = '';
    
    //Lookup prams
    if(isset($ipqs['config']['vpnproxy']['enabled'])      === false) $ipqs['config']['vpnproxy']['enabled']     = true;
    if(isset($ipqs['config']['vpnproxy']['fraud'])        === false) $ipqs['config']['vpnproxy']['fraud']       = 75;
    if(isset($ipqs['config']['vpnproxy']['block'])        === false) $ipqs['config']['vpnproxy']['block']       = '';
    if(isset($ipqs['config']['vpnproxy']['cache'])        === false) $ipqs['config']['vpnproxy']['cache']       = '1h';
    if(isset($ipqs['config']['vpnproxy']['addr'])         === false) $ipqs['config']['vpnproxy']['addr']        = [];
    if(isset($ipqs['config']['vpnproxy']['countires'])    === false) $ipqs['config']['vpnproxy']['countires']   = '';
    if(isset($ipqs['config']['vpnproxy']['strictness'])   === false) $ipqs['config']['vpnproxy']['strictness']  = '1';
    if(isset($ipqs['config']['vpnproxy']['publicap'])     === false) $ipqs['config']['vpnproxy']['publicap']    = false;
};

/**
 * Helper to create the nav menu for our module.
 *
 * @return void
 */
$ipqs['fnc']['nav'] = function(){
    $page = $_SERVER['PHP_SELF'];
    $tab_array	= array();
    $tab_array[]	= array(gettext('General'),	($page === '/ipqualityscore/index.php'? true:false),	'/ipqualityscore/index.php');
    $tab_array[]	= array(gettext('Lookup'),	($page === '/ipqualityscore/lookup.php'? true:false),	'/ipqualityscore/lookup.php');
    $tab_array[]	= array(gettext('Statistics'),	($page === '/ipqualityscore/stats.php'? true:false),	'/ipqualityscore/stats.php');
    return $tab_array;
};

/**
 * Simple parser to make stuff looking nice.
 */
$ipqs['fnc']['json_to_str'] = function($json){
    $str ='';
    foreach(json_decode($json) as $k => $v){
        $str = $str . $k . " => " . $v . "\n";
    }
    return $str;
};

/**
 * Function to setup oauth within the config.xml.
 */
$ipqs['fnc']['set_oauth'] = function($domain, $access_token){
    global $config, $ipqs;
    $ipqs['config']		= &$config['installedpackages']['ipqualityscore']['config'][0];
    $ipqs['config']['oauth']['access_token'] = $access_token;
    $ipqs['config']['oauth']['domain'] = $domain;
    write_config('[IPQualityScore] - updated oauth');
};

/**
 * Function removes oauth key and domain from config.xml
 */
$ipqs['fnc']['del_oauth'] = function(){
    global $config, $ipqs;
    $ipqs['config']		= &$config['installedpackages']['ipqualityscore']['config'][0];
    $ipqs['config']['oauth']['access_token'] = '';
    $ipqs['config']['oauth']['domain'] = '';
    write_config('[IPQualityScore] - cleard oauth');
};


/**
 * Check address against IPQS.
 */
$ipqs['fnc']['validateaddress'] = function($ip, $access_token, $domain, $strictness = 1, $publicap = false){
    $url = 'https://ipqualityscore.com/webhooks/pfsense/json/proxy_check';
    $param = array(
        'secret' => $access_token,
        'key' => $domain,
        'ip' => $ip,
        'allow_public_access_points' => $ap,
        'transaction_strictness' => $publicap,

    );
    $curl = curl_init();
    curl_setopt($curl, CURLOPT_URL, $url);
    curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($curl, CURLOPT_FOLLOWLOCATION, true);
    curl_setopt($curl, CURLOPT_POST, 1);
    curl_setopt($curl, CURLOPT_POSTFIELDS, http_build_query($param));
    $json = curl_exec($curl);
    $response_code = curl_getinfo($curl);
    curl_close($curl);

    return $json;
};

/**
 * Block address, add to alias.
 */
$ipqs['fnc']['blockaddr'] = function($ip){
    global $config, $ipqs;
    $config = parse_config(true);
    $ipqs['config']		= &$config['installedpackages']['ipqualityscore']['config'][0];
    $current_ts = date_timestamp_get(date_create());
    //echo 'addr_'.$ip;

    if(!isset($ipqs['config']['vpnproxy']['addr']['addr_'.$ip])){
        $ipqs['config']['vpnproxy']['addr'] = array_merge ($ipqs['config']['vpnproxy']['addr'] , ['addr_'.$ip=>$current_ts]);
    }else{
        if(!$ipqs['fnc']['is_cached']($ip)){
            $ipqs['config']['vpnproxy']['addr']['addr_'.$ip] = date_timestamp_get(date_create());  
        }
    }

    $ipqs['rule']['add']($ip);
    return;
};

/**
 * Adds ip to cahce insure there are no duplicate entries.
 */
$ipqs['fnc']['add_cache'] = function($ip){
    global $config, $ipqs;
    $config = parse_config(true);
    $ipqs['config'] = &$config['installedpackages']['ipqualityscore']['config'][0];

    if(!is_array($ipqs['config']['vpnproxy']['addr'])){
        $ipqs['config']['vpnproxy']['addr']=Array();
    }

    if(!isset($ipqs['config']['vpnproxy']['addr']['addr_'.$ip])){
        $ipqs['config']['vpnproxy']['addr'] = array_merge ($ipqs['config']['vpnproxy']['addr'] , ['addr_'.$ip=>$current_ts]);
    }else{

        $cached_timestamp = $ipqs['config']['vpnproxy']['addr']['addr_'.$ip];
        $ts = date_timestamp_get(date_create());
        switch ($ipqs['config']['vpnproxy']['cache']){
            case '1h':
                $cached_timestamp += 3600;
                break;
            case '3h':
                $cached_timestamp += 10800;
                break;
            case '6h':
                $cached_timestamp += 21600;
                break;
            case '12h':
                $cached_timestamp += 43200;
                break;
            case '1d':
                $cached_timestamp += 86400;
                break;
        };

        //insures that they checks happen when they should.
        if($ts > $cached_timestamp){
            $ipqs['config']['vpnproxy']['addr']['addr_'.$ip] = date_timestamp_get(date_create());  
        }
    }
};

$ipqs['fnc']['is_cached'] = function($ip){
    global $config, $ipqs;
    $ipqs['config']	= &$config['installedpackages']['ipqualityscore']['config'][0];
    $cached_timestamp = $ipqs['config']['vpnproxy']['addr']['addr_'.$ip];

    switch ($ipqs['config']['vpnproxy']['cache']){
        case '1h':
            $cached_timestamp += 3600;
            break;
        case '3h':
            $cached_timestamp += 10800;
            break;
        case '6h':
            $cached_timestamp += 21600;
            break;
        case '12h':
            $cached_timestamp += 43200;
            break;
        case '1d':
            $cached_timestamp += 86400;
            break;
    };

    echo $cached_timestamp . "=> " . date_timestamp_get(date_create()) ."\n";
    if($cached_timestamp > date_timestamp_get(date_create())){
        
        return true;
    }

    return false;
};
/**
 * Used to clean up address in the stacks for block lists.
 */
$ipqs['fnc']['rmaddr'] = function($ip = null){
    global  $ipqs;

    if($ip === null){
        $current_ts = date_timestamp_get(date_create());
        $cache_period = 0;

        switch ($ipqs['config']['vpnproxy']['cache']){
            case '1h':
                $cache_period += 3600;
                break;
            case '3h':
                $cache_period += 10800;
                break;
            case '6h':
                $cache_period += 21600;
                break;
            case '12h':
                $cache_period += 43200;
                break;
            case '1d':
                $cache_period += 86400;
                break;
        }

        foreach($ipqs['config']['vpnproxy']['addr'] as $addr => $timecached){
            if($timecached+ $cache_period < $current_ts){
                unset($ipqs['config']['vpnproxy']['addr']['addr_'.$addr]);
                $ipqs['rule']['remove']($addr);
            }
        }


        return;
    }

    if(isset($ipqs['config']['vpnproxy']['addr']['addr_'.$ip])){
        unset($ipqs['config']['vpnproxy']['addr']['addr_'.$ip]);
        $ipqs['rule']['remove']($ip);
    }
};

//Installer function

$ipqs['installer']['fnc']['install'] = function(){
    global $config;
    $arr = Array();
    $fpkg = false; // to add packages info.
    $fmnu = false; // to add menu.
    $fals = false; // to add alias.

    echo "\e[32mStarting IPQualityscore Installer\n\n";

    if(!isset($config['installedpackages']['package'])){
        $config['installedpackages'] = array_merge($config['installedpackages'],['package'=>[]]);
    }

    foreach($config['installedpackages']['package'] as $pkg){
        if($pkg['name'] == "ipqualityscore"){
            $fpgk = true;
            break;
        }

    }

    if($fpgk == false){
        echo "\e[34m Addding package to list.\n";
        $arr = [count($config['installedpackages']['package']) =>[
            'name' => 'ipqualityscore',
            'pkginfolink' => 'https://www.ipqualityscore.com/documentation/integrations/overview',
            'descr' => 'fraud prevention tools',
            'websote' => 'https://www.ipqualityscore.com/',
            'version' => '0.1.0',
            'include_file' => '/usr/local/pkg/ipqualityscore/ipqualityscore.xml'
        ]];
        $config['installedpackages']['package'] = array_merge($config['installedpackages']['package'], $arr);
        unset($arr);
    }


    if(!isset($config['installedpackages']['menu'])){
        $config['installedpackages'] = array_merge($config['installedpackages'],['menu'=>[]]);
    }

    foreach($config['installedpackages']['menu'] as $mnu){
        if($mnu['name'] == 'IPQualityScore'){
            $fmnu= true;
            break;
        }
    }

    if($fmnu == false){
        echo "\e[34m Addding package to menu.\n";
        $arr = [ count($config['installedpackages']['menu'])=>[
            'name' => 'IPQualityScore',
            'tooltiptext' =>  '',
            'section' => 'Firewall',
            'url' => '/ipqualityscore/index.php',
        ]
        ];

        $config['installedpackages']['menu'] = array_merge($config['installedpackages']['menu'], $arr);
        unset($arr);

    }

    init_config_arr(array('aliases', 'alias'));

    $ipqs['alias'] =  &$config['aliases']['alias'];

    foreach($config['aliases']['alias'] as $als){
        
        if($als['name'] == 'ipqs_blocklist'){
            $fals = true;
            break;
        }
    }

    if($fals == false){
        
        echo "\e[34m Creating Alias \"ipqs_blocklist\"\n";
        $arr = [
            'name' => 'ipqs_blocklist',
            'type' => 'host',
            'address' => " ",
            'descr' => 'managed by IPQS',
            'details' => 'DO NOT CHNAGE',
        ];

        $input_errors = saveAlias($arr,NULL);
        mark_subsystem_dirty('aliases');
        unset($arr);
    }
    
    echo "\e[34m Setting variables.\n";
    $arr = ['ipqualityscore' => [
        'config'=>[
            0 => [
                'gbl' => [
                    'authorized'    => '127.0.0.1'
                ],
                'oauth' =>[
                    'domain'        => '',
                    'access_token'  => '',
                ],
                'vpnproxy' => [
                    'enabled'       => false,
                    'fraud'         => '75',
                    'block'         => '',
                    'cache'         => '1h',
                    'addr'          => '',
                    'countires'     => '',
                    'strictness'    => '',
                    'publicap'      => '',
                ]
            ]
        ]
    ]];
    $config['installedpackages'] = array_merge($config['installedpackages'],$arr);
    unset($arr);


    echo "\n\n\e[32mInstall Completed... \e[39m \n\n";
    write_config('[IPQualityScore] - created alias "ipqs_blocklist".');
};

/**
 * uninstall sequence.
 */
$ipqs['installer']['fnc']['uninstall'] = function(){
    global $config;
    unset($config['installedpackages']['ipqualityscore']);

    //check if we have menus to look at.
    if($config['installedpackages']['menu']){
        for( $i = 0; $i === count($config['installedpackages']['menu']) -1; $i++){
            if($config['installedpackages']['menu'][$i]['name'] == "IPQualityScore"){
                unset($config['installedpackages']['menu'][$i]);
            }
        }
    }

    if($config['installedpackages']['package']){
        for( $i = 0; $i === count($config['installedpackages']['package']) -1; $i++){
            if($config['installedpackages']['package'][$i]['name'] == "ipqualityscore"){
                unset($config['installedpackages']['menu'][$i]);
            }
        }
    }

    write_config("removing IPQualityScore");
    exec("ps aux | grep \"/usr/local/bin/ipqualityscore-netmon\" | awk '{print $2}' | xargs kill -9");
    exec("sleep 5 && rm -rf /usr/local/www/ipqualityscore && rm -rf /usr/local/pkg/ipqualityscore && rm -rf /usr/local/pkg/ipqualityscore.xml && rm -rf /usr/local/bin/ipqualityscore-netmon &");
};

/**
 * addes ip to block list
 */
$ipqs['rule']['add'] = function($ip){
    global $config, $ipqs;
    $ipqs['config']		= &$config['installedpackages']['ipqulaityscore']['config'][0];
    $ipqs['alias'] = &$config['aliases']['alias'];
    $bFlag = false;
    $alias = Array();
    foreach($ipqs['alias'] as &$a){
        if($a['name'] === "ipqs_blocklist"){
            $bFlag = true;
            $alias = &$a;
            break;
        }
    }

    if($bFlag){
        
        if(strpos($alias['address'], $ip) === false){
            
            $alias['address'] = $alias['address'] .(strlen($alias['address']) == 0?'':' ') . trim($ip);
        } 
    }else{
        $alias = Array(
            'name'      => 'ipqs_blocklist',
            'type'      => 'host',
            'address'   => trim($ip),
            'descr'     => 'Managed by IPQS do not touch',
            'details'   => ''
        );
        array_push($ipqs['alias'], $alias);
    }

    
    write_config('added address to block list');
};

/**
 * pulling ip from blocklist.
 */
$ipqs['rule']['remove'] = function($ip){
    global $config, $ipqs;
    $ipqs['config']		= &$config['installedpackages']['ipqulaityscore']['config'][0];
    $ipqs['alias'] = &$config['aliases']['alias'];


    $bFlag = false;
    $alias = Array();

    foreach($ipqs['alias'] as &$a){
        if($a['name'] === "ipqs_blocklist"){
            $bFlag = true;
            $alias = $a;
            break;
        }
    }

    if($bFlag){
        $pos = strpos($alias['address'], $ip);
        if($pos !== flase){
         $arr  = explode(" ",$alias['address']);
         unset($arr[$ip]);
         $alias['address'] = implode(" ", $arr);
         write_config('removed address to block list');
        } 
     }
};

$ipqs['bin']['start'] = function(){
    exec('pgrep ipqualityscore-netm', $pid);
    if(!empty($pid)) return;
    exec('/usr/local/bin/ipqualityscore-netmon  > /dev/null &');
};

$ipqs['bin']['stop'] = function(){
    exec("ps aux | grep \"/usr/local/bin/ipqualityscore-netmon\" | awk '{print $2}' | xargs kill -9");
};

$ipqs['bin']['watchdog'] = function(){
    exec('pgrep ipqualityscore-netm', $pid);
    if(empty($pid)){
        $ipqs['bin']['start'];
    }
};


$ipqs['vars']['countires']=Array(
    'NONE'=>"None",
    'AF' =>"Afghanistan",
    'AX' =>"Åland Islands",
    'AL' =>"Albania",
    'DZ' =>"Algeria",
    'AS' =>"American Samoa",
    'AD' =>"Andorra",
    'AO' =>"Angola",
    'AI' =>"Anguilla",
    'AQ' =>"Antarctica",
    'AG' =>"Antigua and Barbuda",
    'AR' =>"Argentina",
    'AM' =>"Armenia",
    'AW' =>"Aruba",
    'AU' =>"Australia",
    'AT' =>"Austria",
    'AZ' =>"Azerbaijan",
    'BS' =>"Bahamas (the)",
    'BH' =>"Bahrain",
    'BD' =>"Bangladesh",
    'BB' =>"Barbados",
    'BY' =>"Belarus",
    'BE' =>"Belgium",
    'BZ' =>"Belize",
    'BJ' =>"Benin",
    'BM' =>"Bermuda",
    'BT' =>"Bhutan",
    'BO' =>"Bolivia (Plurinational State of)",
    'BQ' =>"Bonaire, Sint Eustatius and Saba",
    'BA' =>"Bosnia and Herzegovina",
    'BW' =>"Botswana",
    'BV' =>"Bouvet Island",
    'BR' =>"Brazil",
    'IO' =>"British Indian Ocean Territory (the)",
    'BN' =>"Brunei Darussalam",
    'BG' =>"Bulgaria",
    'BF' =>"Burkina Faso",
    'BI' =>"Burundi",
    'CV' =>"Cabo Verde",
    'KH' =>"Cambodia",
    'CM' =>"Cameroon",
    'CA' =>"Canada",
    'KY' =>"Cayman Islands (the)",
    'CF' =>"Central African Republic (the)",
    'TD' =>"Chad",
    'CL' =>"Chile",
    'CN' =>"China",
    'CX' =>"Christmas Island",
    'CC' =>"Cocos (Keeling) Islands (the)",
    'CO' =>"Colombia",
    'KM' =>"Comoros (the)",
    'CD' =>"Congo (the Democratic Republic of the)",
    'CG' =>"Congo (the)",
    'CK' =>"Cook Islands (the)",
    'CR' =>"Costa Rica",
    'HR' =>"Croatia",
    'CU' =>"Cuba",
    'CW' =>"Curaçao",
    'CY' =>"Cyprus",
    'CZ' =>"Czechia",
    'CI' =>"Côte d'Ivoire",
    'DK' =>"Denmark",
    'DJ' =>"Djibouti",
    'DM' =>"Dominica",
    'DO' =>"Dominican Republic (the)",
    'EC' =>"Ecuador",
    'EG' =>"Egypt",
    'SV' =>"El Salvador",
    'GQ' =>"Equatorial Guinea",
    'ER' =>"Eritrea",
    'EE' =>"Estonia",
    'SZ' =>"Eswatini",
    'ET' =>"Ethiopia",
    'FK' =>"Falkland Islands (the) [Malvinas]",
    'FO' =>"Faroe Islands (the)",
    'FJ' =>"Fiji",
    'FI' =>"Finland",
    'FR' =>"France",
    'GF' =>"French Guiana",
    'PF' =>"French Polynesia",
    'TF' =>"French Southern Territories (the)",
    'GA' =>"Gabon",
    'GM' =>"Gambia (the)",
    'GE' =>"Georgia",
    'DE' =>"Germany",
    'GH' =>"Ghana",
    'GI' =>"Gibraltar",
    'GR' =>"Greece",
    'GL' =>"Greenland",
    'GD' =>"Grenada",
    'GP' =>"Guadeloupe",
    'GU' =>"Guam",
    'GT' =>"Guatemala",
    'GG' =>"Guernsey",
    'GN' =>"Guinea",
    'GW' =>"Guinea-Bissau",
    'GY' =>"Guyana",
    'HT' =>"Haiti",
    'HM' =>"Heard Island and McDonald Islands",
    'VA' =>"Holy See (the)",
    'HN' =>"Honduras",
    'HK' =>"Hong Kong",
    'HU' =>"Hungary",
    'IS' =>"Iceland",
    'IN' =>"India",
    'ID' =>"Indonesia",
    'IR' =>"Iran (Islamic Republic of)",
    'IQ' =>"Iraq",
    'IE' =>"Ireland",
    'IM' =>"Isle of Man",
    'IL' =>"Israel",
    'IT' =>"Italy",
    'JM' =>"Jamaica",
    'JP' =>"Japan",
    'JE' =>"Jersey",
    'JO' =>"Jordan",
    'KZ' =>"Kazakhstan",
    'KE' =>"Kenya",
    'KI' =>"Kiribati",
    'KP' =>"Korea (the Democratic People's Republic of)",
    'KR' =>"Korea (the Republic of)",
    'KW' =>"Kuwait",
    'KG' =>"Kyrgyzstan",
    'LA' =>"Lao People's Democratic Republic (the)",
    'LV' =>"Latvia",
    'LB' =>"Lebanon",
    'LS' =>"Lesotho",
    'LR' =>"Liberia",
    'LY' =>"Libya",
    'LI' =>"Liechtenstein",
    'LT' =>"Lithuania",
    'LU' =>"Luxembourg",
    'MO' =>"Macao",
    'MG' =>"Madagascar",
    'MW' =>"Malawi",
    'MY' =>"Malaysia",
    'MV' =>"Maldives",
    'ML' =>"Mali",
    'MT' =>"Malta",
    'MH' =>"Marshall Islands (the)",
    'MQ' =>"Martinique",
    'MR' =>"Mauritania",
    'MU' =>"Mauritius",
    'YT' =>"Mayotte",
    'MX' =>"Mexico",
    'FM' =>"Micronesia (Federated States of)",
    'MD' =>"Moldova (the Republic of)",
    'MC' =>"Monaco",
    'MN' =>"Mongolia",
    'ME' =>"Montenegro",
    'MS' =>"Montserrat",
    'MA' =>"Morocco",
    'MZ' =>"Mozambique",
    'MM' =>"Myanmar",
    'NA' =>"Namibia",
    'NR' =>"Nauru",
    'NP' =>"Nepal",
    'NL' =>"Netherlands (the)",
    'NC' =>"New Caledonia",
    'NZ' =>"New Zealand",
    'NI' =>"Nicaragua",
    'NE' =>"Niger (the)",
    'NG' =>"Nigeria",
    'NU' =>"Niue",
    'NF' =>"Norfolk Island",
    'MP' =>"Northern Mariana Islands (the)",
    'NO' =>"Norway",
    'OM' =>"Oman",
    'PK' =>"Pakistan",
    'PW' =>"Palau",
    'PS' =>"Palestine, State of",
    'PA' =>"Panama",
    'PG' =>"Papua New Guinea",
    'PY' =>"Paraguay",
    'PE' =>"Peru",
    'PH' =>"Philippines (the)",
    'PN' =>"Pitcairn",
    'PL' =>"Poland",
    'PT' =>"Portugal",
    'PR' =>"Puerto Rico",
    'QA' =>"Qatar",
    'MK' =>"Republic of North Macedonia",
    'RO' =>"Romania",
    'RU' =>"Russian Federation (the)",
    'RW' =>"Rwanda",
    'RE' =>"Réunion",
    'BL' =>"Saint Barthélemy",
    'SH' =>"Saint Helena, Ascension and Tristan da Cunha",
    'KN' =>"Saint Kitts and Nevis",
    'LC' =>"Saint Lucia",
    'MF' =>"Saint Martin (French part)",
    'PM' =>"Saint Pierre and Miquelon",
    'VC' =>"Saint Vincent and the Grenadines",
    'WS' =>"Samoa",
    'SM' =>"San Marino",
    'ST' =>"Sao Tome and Principe",
    'SA' =>"Saudi Arabia",
    'SN' =>"Senegal",
    'RS' =>"Serbia",
    'SC' =>"Seychelles",
    'SL' =>"Sierra Leone",
    'SG' =>"Singapore",
    'SX' =>"Sint Maarten (Dutch part)",
    'SK' =>"Slovakia",
    'SI' =>"Slovenia",
    'SB' =>"Solomon Islands",
    'SO' =>"Somalia",
    'ZA' =>"South Africa",
    'GS' =>"South Georgia and the South Sandwich Islands",
    'SS' =>"South Sudan",
    'ES' =>"Spain",
    'LK' =>"Sri Lanka",
    'SD' =>"Sudan (the)",
    'SR' =>"Suriname",
    'SJ' =>"Svalbard and Jan Mayen",
    'SE' =>"Sweden",
    'CH' =>"Switzerland",
    'SY' =>"Syrian Arab Republic",
    'TW' =>"Taiwan (Province of China)",
    'TJ' =>"Tajikistan",
    'TZ' =>"Tanzania, United Republic of",
    'TH' =>"Thailand",
    'TL' =>"Timor-Leste",
    'TG' =>"Togo",
    'TK' =>"Tokelau",
    'TO' =>"Tonga",
    'TT' =>"Trinidad and Tobago",
    'TN' =>"Tunisia",
    'TR' =>"Turkey",
    'TM' =>"Turkmenistan",
    'TC' =>"Turks and Caicos Islands (the)",
    'TV' =>"Tuvalu",
    'UG' =>"Uganda",
    'UA' =>"Ukraine",
    'AE' =>"United Arab Emirates (the)",
    'GB' =>"United Kingdom of Great Britain and Northern Ireland (the)",
    'UM' =>"United States Minor Outlying Islands (the)",
    'US' =>"United States of America (the)",
    'UY' =>"Uruguay",
    'UZ' =>"Uzbekistan",
    'VU' =>"Vanuatu",
    'VE' =>"Venezuela (Bolivarian Republic of)",
    'VN' =>"Viet Nam",
    'VG' =>"Virgin Islands (British)",
    'VI' =>"Virgin Islands (U.S.)",
    'WF' =>"Wallis and Futuna",
    'EH' =>"Western Sahara",
    'YE' =>"Yemen",
    'ZM' =>"Zambia",
    'ZW' =>"Zimbabwe",
);

